/*****************************************************************************/
/*                                                                           */
/*   Component   : Sample Port Connection Manager (PCM)                      */
/*                                                                           */
/*   Module      : PCMCALL.C                                                 */
/*                                                                           */
/*   Description : This module contains the routines for the GCCI            */
/*                 primitives associated with call instances.                */
/*                                                                           */
/*   Property of IBM.                                                        */
/*   This sample code is available without warranty upon request.            */
/*                                                                           */
/*****************************************************************************/

#define INCL_DOS
#include <os2.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "gcci_com.h"
#include "gcci_cm.h"
#include "cci_rc.h"
#include "sf.h"
#include "pcmdef.h"
#include "pcmproto.h"



/*****************************************************************************/
/*                                                                           */
/*   Routine     : CallSetupRequest                                          */
/*                                                                           */
/*   Description : This routine processes a Call Setup request primitive.    */
/*                                                                           */
/*                 This primitive is used to initiate an outgoing call.      */
/*                 This routine creates a call instance and processes        */
/*                 the subfields supplied with the primitive.  A Call Setup  */
/*                 confirm primitive is sent to the Connection Manager to    */
/*                 indicate that this primitive is accepted and that the     */
/*                 call attempt is under way.                                */
/*                                                                           */
/*****************************************************************************/

VOID near CallSetupRequest( CI_PCM_CALL_SETUP_REQ *pRequest )
{
   PSAP   pSAP;
   PPCMCB pPCMCB;
   PCALL  pCall;
   USHORT RetCode;

   pSAP   = (PSAP)pRequest->P_SAP_ID;
   pPCMCB = pSAP->pPCMCB;

   /* Create a Call Instance */
   pCall = CreateCall( pPCMCB );

   if ( pCall == 0 )
      {
         CallSetupConfirm( pPCMCB, pRequest->U_CI_ID, RetCode );
         return;
      }

   /* Store the U_CI_ID */
   pCall->U_CI_ID = pRequest->U_CI_ID;

   /* Process the subfields for this primitive */
   RetCode = ProcessSubfields( pRequest );
   if ( RetCode != CCI_NO_ERROR )
      {
         CallSetupConfirm( pPCMCB, pRequest->U_CI_ID, RetCode );
         DeleteCall( pCall );
         return;
      }

   /* Send a Call Setup confirm primitive to the CM */
   if ( CallSetupConfirm(pPCMCB,pRequest->U_CI_ID,CCI_NO_ERROR) == SUCCESS )
         ActivateConn( pCall );
   else  /* Error sending the Call Setup Confirm */
         DeleteCall( pCall );
   return;
}



/*****************************************************************************/
/*                                                                           */
/*   Routine     : CreateCall                                                */
/*                                                                           */
/*   Description : This routine creates a new Call structure,                */
/*                 initializes it, and adds it to the end of the list        */
/*                 of Calls.                                                 */
/*                                                                           */
/*****************************************************************************/

PCALL near CreateCall( PPCMCB pPCMCB )
{
   PCALL pCall, pLastCall;

   /* Create a new Call Instance */
   pCall = malloc( sizeof(CALL) );
   if ( pCall == 0 )
      {
         return pCall;
      }

   /* Initialize the Port variables */
   pCall->pNextCall   = NULL;
   pCall->pPCMCB      = pPCMCB;

   /* Add the new Port to the Port list */
   if ( pPCMCB->pCallList == NULL )
      pPCMCB->pCallList = pCall;
   else
      {
         for ( pLastCall=pPCMCB->pCallList;
               pLastCall->pNextCall!=NULL;
               pLastCall=pLastCall->pNextCall );
         pLastCall->pNextCall = pCall;
      }

   return pCall;
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : DeleteCall                                                */
/*                                                                           */
/*   Description : This routine removes a Call structure from the            */
/*                 Call list and releases its memory.                        */
/*                                                                           */
/*****************************************************************************/

PCALL near DeleteCall( PCALL pCall)
{
   PCALL pPrevCall, pNextCall;
   PPCMCB pPCMCB;

   pPCMCB = pCall->pPCMCB;

   /* Remove the Call from the Call list */
   if ( pPCMCB->pCallList == pCall )
      {
         pNextCall = pPCMCB->pCallList = pCall->pNextCall;
      }
   else
      {
         for ( pPrevCall=pPCMCB->pCallList;
               pPrevCall->pNextCall!=pCall;
               pPrevCall=pPrevCall->pNextCall );
         pNextCall = pPrevCall->pNextCall = pCall->pNextCall;
      }

   free( pCall );   /* Release memory for the Port structure */

   return pNextCall;
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : ProcessSubfields                                          */
/*                                                                           */
/*   Description : This routine processes the subfields of a Call Setup      */
/*                 request primitive.  Subfieds are routed to the            */
/*                 appropriate processing routines.                          */
/*                                                                           */
/*****************************************************************************/

USHORT near ProcessSubfields( CI_PCM_CALL_SETUP_REQ *pRequest )
{
   SF_CC_BEAR_CAP *pBearer;
   SF_CC_BEAR_CAP *pDataInfoLimit;
   USHORT RetCode;

   /* Setup pBearer to point to the first subfield */
   SELECTOROF(pBearer) = SELECTOROF(pRequest);
   OFFSETOF(pBearer)   = OFFSETOF(pRequest) + pRequest->CtrlInfoLength;

   /* Setup pDataInfoLimit to point just beyond the last subfield */
   SELECTOROF(pDataInfoLimit) = SELECTOROF(pRequest);
   OFFSETOF(pDataInfoLimit)   = OFFSETOF(pBearer) + pRequest->DataInfoLength;

   while( pBearer < pDataInfoLimit )
      {
         switch( pBearer->SFType )
            {
               case SF_CC_TYPE_BEAR_CAP:
                  RetCode = ProcessBearerCap( pBearer);
                  if ( RetCode != CCI_NO_ERROR ) return RetCode;
                  break;
               case SF_CC_TYPE_CALLED_NUMBER:
                  RetCode = ProcessCalledNum( (SF_CC_CALLED_NUMBER *)pBearer);
                  if ( RetCode != CCI_NO_ERROR ) return RetCode;
                  break;
               case SF_CS_TYPE_PERMANENT:
                  RetCode = ProcessPermConn( (SF_CS_PERMANENT *)pBearer);
                  if ( RetCode != CCI_NO_ERROR ) return RetCode;
                  break;
            }

         /* Go to the next subfield */
         OFFSETOF(pBearer) += pBearer->SFLength;
      }

   /* Note: Other procedures may be used to determine the port, channel, or other      */
   /*       line for connecting a call instance.  The "port" must be recorded and      */
   /*       connected to fulfill the Call Setup Request.  These routines are           */
   /*       provided as placeholders to suggest the work to be done by PCM developers  */
   /*       and not as a skeleton for that development, though it may be used in that  */
   /*       way if desired.                                                            */

   return CCI_NO_ERROR;
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : ProcessBearerCap                                          */
/*                                                                           */
/*   Description : This routine processes the Bearer Capability              */
/*                 subfield of a Call Setup request primitive.               */
/*                                                                           */
/*****************************************************************************/

USHORT near ProcessBearerCap( SF_CC_BEAR_CAP *pBearer )
{
   return CCI_NO_ERROR;
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : ProcessCalledNum                                          */
/*                                                                           */
/*   Description : This routine processes the Called Party Number            */
/*                 subfield of a Call Setup request primitive.               */
/*                                                                           */
/*****************************************************************************/

USHORT near ProcessCalledNum( SF_CC_CALLED_NUMBER *pCalledNum )
{

   return CCI_NO_ERROR;
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : ProcessCallingNum                                         */
/*                                                                           */
/*   Description : This routine processes the Calling Party Number           */
/*                 subfield of a Call Setup request primitive.               */
/*                                                                           */
/*****************************************************************************/

USHORT near ProcessCallingNum( SF_CC_CALLING_NUMBER *pCallingNum, PPCMCB pPCMCB )
{
   return CCI_NO_ERROR;
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : ProcessChanID                                             */
/*                                                                           */
/*   Description : This routine processes the Channel Identification         */
/*                 subfield of a Call Setup request primitive.               */
/*                                                                           */
/*****************************************************************************/

USHORT near ProcessChanID( SF_CC_CHANNEL_ID *pChanID, PPCMCB pPCMCB )
{

   return CCI_NO_ERROR;
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : ProcessPermConn                                          */
/*                                                                           */
/*   Description : This routine processes the Permanent Connection           */
/*                 subfield of a Call Setup request primitive for the        */
/*                 purpose of setting answer role parameters.                */
/*                                                                           */
/*****************************************************************************/

USHORT near ProcessPermConn( SF_CS_PERMANENT *pPermConn )
{
   return CCI_NO_ERROR;
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : CallSetupConfirm                                          */
/*                                                                           */
/*   Description : This routine sends a Call Setup confrim primitive         */
/*                 with the specified completion code to the Connection      */
/*                 Manager.  If the completion code is successful, then      */
/*                 a Channel Identification subfield and a Call Status       */
/*                 subfield are included.  The Channel Indentification       */
/*                 subfield informs the Connection Manager which channel     */
/*                 is being used for the call attempt.                       */
/*                                                                           */
/*****************************************************************************/

BOOL near CallSetupConfirm( PPCMCB pPCMCB, ULONG U_CI_ID, USHORT CompCode )
{
   CI_PCM_CALL_SETUP_CNF *pConfirm;
   SF_CC_CHANNEL_ID      *pChanID;
   SF_CS_CALL_STATUS     *pCallStat;
   USHORT RetCode, ConfirmSize, x;

   ConfirmSize = sizeof(CI_PCM_CALL_SETUP_CNF) + sizeof(SF_CC_CHANNEL_ID) +
                 sizeof(SF_CS_CALL_STATUS);

   /* Get memory for Call Setup Confirm primitive */
   pConfirm = malloc( ConfirmSize );
   if ( pConfirm == 0 )
      {
         return FAILURE;
      }

   /* Fill in Call Setup Control Information Field */
   pConfirm->CtrlInfoLength = sizeof(CI_PCM_CALL_SETUP_CNF);
   pConfirm->PrimitiveCode  = GCCI_CM_PC_PCM_CALL_SETUP_CNF;
   pConfirm->Selector       = GCCI_COM_SEL_PCM;
   pConfirm->IDType         = GCCI_COM_ID_TYPE_U_CI_ID;
   pConfirm->U_CI_ID        = U_CI_ID;
   pConfirm->CommonStatus.ReportLocation   = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pConfirm->CommonStatus.ReportCompCode   = CompCode;
   pConfirm->CommonStatus.Reserved         = 0;
   pConfirm->CommonStatus.OrigLocation     = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pConfirm->CommonStatus.OrigCompCode     = CompCode;
   pConfirm->CommonStatus.SysLogID         = 0L;

   if ( CompCode == CCI_NO_ERROR )
      {
         pConfirm->CommonStatus.CompCodeCategory = GCCI_COM_CMP_CAT_REQ_SUCCESSFUL;

         pConfirm->DataInfoLength = sizeof(SF_CC_CHANNEL_ID) +
                                    sizeof(SF_CS_CALL_STATUS);

         /* Fill in Channel ID Subfield */
         SELECTOROF(pChanID)    = SELECTOROF(pConfirm);
         OFFSETOF(pChanID)      = OFFSETOF(pConfirm)+sizeof(CI_PCM_CALL_SETUP_CNF);
         pChanID->SFLength      = sizeof(SF_CC_CHANNEL_ID);
         pChanID->SFIdentifier  = SF_SFID_CCITT;
         pChanID->SFType        = SF_CC_TYPE_CHANNEL_ID;
         pChanID->SFChannelPref = SF_CHANNEL_REQUIRED;
         pChanID->SFDChannelInd = SF_B_CHANNEL_IDENTIFIED;
         pChanID->SFChannelSel  = SF_CHANNEL_IN_SLOT_MAP;

         /* Fill in Channel Map */
         for ( x=0; x<12; x++ ) pChanID->SFChanMap[x] = 0;

         /* Fill in Call Status Subfield */
         SELECTOROF(pCallStat)    = SELECTOROF(pConfirm);
         OFFSETOF(pCallStat)      = OFFSETOF(pChanID)+sizeof(SF_CC_CHANNEL_ID);
         pCallStat->SFLength      = sizeof(SF_CS_CALL_STATUS);
         pCallStat->SFIdentifier  = SF_SFID_CALL_CONTROL;
         pCallStat->SFType        = SF_CS_TYPE_CALL_STATUS;
         pCallStat->SFCallStatus  = SF_STATUS_CALL_PROC;
         pCallStat->SFReserved1   = 0;
      }
   else  /* CompCode is an error */
      {
         pConfirm->CommonStatus.CompCodeCategory = GCCI_COM_CMP_CAT_REQ_REJECTED;
         pConfirm->DataInfoLength = 0;
      }

   /* Call Connection Manager with Call Setup Confirm primitive */
   RetCode = pPCMCB->CMSAP.Confirm( pConfirm );

   if ( RetCode == CCI_NO_ERROR )
      {
         free( pConfirm );
         return SUCCESS;
      }
   else  /* Error returned on Call Setup Confirm */
      {
         free( pConfirm );
         return FAILURE;
      }
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : CallSetupIndicate                                         */
/*                                                                           */
/*   Description : This routine sends a Call Setup indicate primitive to     */
/*                 the Connection Manager.  This subfield informs the        */
/*                 Connection Manager that an incoming call has been         */
/*                 received.                                                 */
/*                                                                           */
/*****************************************************************************/

BOOL near CallSetupIndicate( PCALL pIncomingCall )
{
   CI_PCM_CALL_SETUP_IND *pIndicate;
   SF_CC_BEAR_CAP        *pBearer;
   SF_CC_CHANNEL_ID      *pChanID;
   SF_CC_CALLED_NUMBER   *pCalledNum;
   USHORT LocalNumLength, CalledSFLength, DataInfoLength;
   USHORT RetCode, x;
   PPCMCB pPCMCB;

   pPCMCB = pIncomingCall->pPCMCB;

   CalledSFLength = sizeof(SF_CC_CALLED_NUMBER);
   DataInfoLength = sizeof(SF_CC_BEAR_CAP) +
                    sizeof(SF_CC_CHANNEL_ID) +
                    CalledSFLength;

   /* Get memory for Call Setup Indicate primitive */
   pIndicate = malloc( sizeof(CI_PCM_CALL_SETUP_IND)+DataInfoLength );
   if ( pIndicate == 0 )
      {
         return FAILURE;
      }

   /* Fill in Call Setup Indicate Control Information Field */
   pIndicate->CtrlInfoLength = sizeof(CI_PCM_CALL_SETUP_IND);
   pIndicate->PrimitiveCode  = GCCI_CM_PC_PCM_CALL_SETUP_IND;
   pIndicate->DataInfoLength = DataInfoLength;
   pIndicate->Selector       = GCCI_COM_SEL_PCM;
   pIndicate->IDType         = GCCI_COM_ID_TYPE_U_SAP_ID;
   pIndicate->U_SAP_ID       = pPCMCB->CMSAP.U_SAP_ID;
   pIndicate->P_CI_ID        = (ULONG)pIncomingCall;

   /* Fill in Bearer Capability Subfield */
   SELECTOROF(pBearer)    = SELECTOROF(pIndicate);
   OFFSETOF(pBearer)      = OFFSETOF(pIndicate)+sizeof(CI_PCM_CALL_SETUP_IND);
   pBearer->SFLength      = sizeof(SF_CC_BEAR_CAP);
   pBearer->SFIdentifier  = SF_SFID_CCITT;
   pBearer->SFType        = SF_CC_TYPE_BEAR_CAP;
   pBearer->SFCallType    = SF_CALL_TYPE_DATA_CIRCUIT;
   pBearer->SFUserRate    = SF_USER_RATE_UNSPECIFIED;
   pBearer->SFUserRateVal = 0L;

   /* Fill in Channel ID Subfield */
   SELECTOROF(pChanID)    = SELECTOROF(pIndicate);
   OFFSETOF(pChanID)      = OFFSETOF(pBearer) + sizeof(SF_CC_BEAR_CAP);
   pChanID->SFLength      = sizeof(SF_CC_CHANNEL_ID);
   pChanID->SFIdentifier  = SF_SFID_CCITT;
   pChanID->SFType        = SF_CC_TYPE_CHANNEL_ID;
   pChanID->SFChannelPref = SF_CHANNEL_REQUIRED;
   pChanID->SFDChannelInd = SF_B_CHANNEL_IDENTIFIED;
   pChanID->SFChannelSel  = SF_CHANNEL_IN_SLOT_MAP;
   pChanID->SFBMacAddr    = 0L;

   /* Fill in Channel Map */
   for ( x=0; x<12; x++ ) pChanID->SFChanMap[x] = 0;

   /* Fill in Called Party Number Subfield */
   SELECTOROF(pCalledNum)     = SELECTOROF(pIndicate);
   OFFSETOF(pCalledNum)       = OFFSETOF(pChanID) + sizeof(SF_CC_CHANNEL_ID);
   pCalledNum->SFLength       = CalledSFLength;
   pCalledNum->SFIdentifier   = SF_SFID_CCITT;
   pCalledNum->SFType         = SF_CC_TYPE_CALLED_NUMBER;
   pCalledNum->SFNumberType   = SF_NUMBER_TYPE_UNKNOWN;
   pCalledNum->SFNumberPlan   = SF_NUMBER_PLAN_UNKNOWN;
   pCalledNum->SFNumberLength = (UCHAR)LocalNumLength;

   /* Call Connection Manager with Call Setup Indicate primitive */
   RetCode = pPCMCB->CMSAP.Indicate( pIndicate );

   if ( RetCode == CCI_NO_ERROR )
      {
         free( pIndicate );
         return SUCCESS;
      }
   else  /* Error returned on Call Setup Indicate */
      {
         free( pIndicate );
         return FAILURE;
      }
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : CallSetupResponse                                         */
/*                                                                           */
/*   Description : This routine processes a Call Setup response primitive.   */
/*                                                                           */
/*****************************************************************************/

VOID near CallSetupResponse( CI_PCM_CALL_SETUP_RSP *pResponse )
{
   PCALL pCall;

   pCall = (PCALL)pResponse->P_CI_ID;

   pCall->U_CI_ID = pResponse->U_CI_ID;


   switch ( pResponse->ActionCode )
      {
         case GCCI_CM_ACT_CODE_ACCEPT_CALL:
            ConnectCall( pCall );
            break;
         case GCCI_CM_ACT_CODE_REJECT_CALL:
            TerminateCall( pCall );
            DeleteCall( pCall );
            break;
         case GCCI_CM_ACT_CODE_PENDING:
            if ( CallClearIndicate( pCall, CCI_PCM_INTERNAL_ERROR ) != SUCCESS )
               {
                 TerminateCall( pCall);
                 DeleteCall( pCall );
               }
            break;
      }
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : CallControlRequest                                        */
/*                                                                           */
/*   Description : This routine processes a Call Control request primitive.  */
/*                                                                           */
/*                 This primitive is not supported by this PCM.              */
/*                                                                           */
/*****************************************************************************/

VOID near CallControlRequest( CI_PCM_CALL_CTRL_REQ *pRequest )
{
   PCALL pCall;

   pCall = (PCALL)pRequest->P_CI_ID;
   CallControlConfirm( pCall, CCI_NO_ERROR);
   return;
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : CallControlConfirm                                        */
/*                                                                           */
/*   Description : This routine sends a Call Control confirm primitive to    */
/*                 the Connection Manager.                                   */
/*                                                                           */
/*                 This primitive is not supported by this PCM.              */
/*                                                                           */
/*****************************************************************************/

BOOL near CallControlConfirm( PCALL pCall, USHORT CompCode )
{
   CI_PCM_CALL_CTRL_CNF *pConfirm;
   PPCMCB pPCMCB;
   USHORT RetCode;

   pPCMCB = pCall->pPCMCB;

   /* Get memory for Call Control Confirm primitive */
   pConfirm = malloc( sizeof(CI_PCM_CALL_CTRL_CNF) );
   if ( pConfirm == 0 )
      {
         return FAILURE;
      }

   /* Fill in Call Control Confirm Control Information Field */
   pConfirm->CtrlInfoLength = sizeof(CI_PCM_CALL_CTRL_CNF);
   pConfirm->PrimitiveCode  = GCCI_CM_PC_PCM_CALL_CTRL_CNF;
   pConfirm->DataInfoLength = 0;
   pConfirm->Selector       = GCCI_COM_SEL_PCM;
   pConfirm->IDType         = GCCI_COM_ID_TYPE_U_CI_ID;
   pConfirm->U_CI_ID        = pCall->U_CI_ID;
   pConfirm->CommonStatus.ReportLocation   = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pConfirm->CommonStatus.Reserved         = 0;
   pConfirm->CommonStatus.OrigLocation     = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pConfirm->CommonStatus.SysLogID         = 0L;

   if ( CompCode == CCI_NO_ERROR )
      pConfirm->CommonStatus.CompCodeCategory = GCCI_COM_CMP_CAT_REQ_SUCCESSFUL;
   else  /* CompCode is an error */
      pConfirm->CommonStatus.CompCodeCategory = GCCI_COM_CMP_CAT_REQ_REJECTED;

   /* Call Connection Manager with Call Control Confirm primitive */
   RetCode = pPCMCB->CMSAP.Confirm( pConfirm );

   if ( RetCode == CCI_NO_ERROR )
      {
         free( pConfirm );
         return SUCCESS;
      }
   else  /* Error returned on Call Control Confirm */
      {
         free( pConfirm );
         return FAILURE;
      }
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : CallClearRequest                                          */
/*                                                                           */
/*   Description : This routine processes a Call Clear request primitive.    */
/*                                                                           */
/*****************************************************************************/

VOID near CallClearRequest( CI_PCM_CALL_CLEAR_REQ *pRequest )
{
   PSAP  pSAP;
   PCALL pCall;

   pSAP = (PSAP)pRequest->Identifier;

   /* Find the call instance */
   for ( pCall=pSAP->pPCMCB->pCallList; pCall->U_CI_ID!=pRequest->U_CI_ID;
         pCall=pCall->pNextCall )
      if ( pCall->pNextCall == NULL )
            return;
   CallClearConfirm( pCall, CCI_NO_ERROR );
   TerminateCall( pCall );
   DeleteCall( pCall );
   return;
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : ActivateConn                                              */
/*                                                                           */
/*   Description : This routine will Activate a channel.                     */
/*                                                                           */
/*****************************************************************************/

VOID near ActivateConn( PCALL pCall )
{
   return;
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : ConnectCall                                               */
/*                                                                           */
/*   Description : This routine will connect a call.                         */
/*                                                                           */
/*****************************************************************************/

VOID near ConnectCall( PCALL pCall )
{
   return;
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : TerminateCall                                             */
/*                                                                           */
/*   Description : This routine will terminate a call.                       */
/*                                                                           */
/*****************************************************************************/

VOID near TerminateCall( PCALL pCall )
{
   return;
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : CallClearConfirm                                          */
/*                                                                           */
/*   Description : This routine sends a Call Clear confirm primitive with    */
/*                 the specified completion code to the Connection Manager.  */
/*                                                                           */
/*****************************************************************************/

BOOL near CallClearConfirm( PCALL pCall, USHORT CompCode )
{
   CI_PCM_CALL_CLEAR_CNF *pConfirm;
   PPCMCB pPCMCB;
   USHORT RetCode;

   pPCMCB = pCall->pPCMCB;

   /* Get memory for Call Clear Confirm primitive */
   pConfirm = malloc( sizeof(CI_PCM_CALL_CLEAR_CNF) );
   if ( pConfirm == 0 )
      {
         return FAILURE;
      }

   /* Fill in Call Clear Confirm Control Information Field */
   pConfirm->CtrlInfoLength = sizeof(CI_PCM_CALL_CLEAR_CNF);
   pConfirm->PrimitiveCode  = GCCI_CM_PC_PCM_CALL_CLEAR_CNF;
   pConfirm->DataInfoLength = 0;
   pConfirm->Selector       = GCCI_COM_SEL_PCM;
   pConfirm->IDType         = GCCI_COM_ID_TYPE_U_CI_ID;
   pConfirm->U_CI_ID        = pCall->U_CI_ID;
   pConfirm->CommonStatus.ReportLocation   = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pConfirm->CommonStatus.Reserved         = 0;
   pConfirm->CommonStatus.OrigLocation     = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pConfirm->CommonStatus.SysLogID         = 0L;

   if ( CompCode == CCI_NO_ERROR )
      pConfirm->CommonStatus.CompCodeCategory = GCCI_COM_CMP_CAT_REQ_SUCCESSFUL;
   else  /* CompCode is an error */
      pConfirm->CommonStatus.CompCodeCategory = GCCI_COM_CMP_CAT_REQ_REJECTED;

   /* Call Connection Manager with Call Clear Confirm primitive */
   RetCode = pPCMCB->CMSAP.Confirm( pConfirm );

   if ( RetCode == CCI_NO_ERROR )
      {
         free( pConfirm );
         return SUCCESS;
      }
   else  /* Error returned on Call Clear Confirm */
      {
         free( pConfirm );
         return FAILURE;
      }
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : CallClearIndicate                                         */
/*                                                                           */
/*   Description : This routine sends a Call Clear indicate primitive with   */
/*                 the specified completion code to the Connection Mangager. */
/*                                                                           */
/*****************************************************************************/

BOOL near CallClearIndicate( PCALL pCall, USHORT CompCode )
{
   CI_PCM_CALL_CLEAR_IND *pIndicate;
   PPCMCB pPCMCB;
   USHORT RetCode;

   pPCMCB = pCall->pPCMCB;

   /* Get memory for Call Clear Indicate primitive */
   pIndicate = malloc( sizeof(CI_PCM_CALL_CLEAR_IND) );
   if ( pIndicate == 0 )
         return FAILURE;

   /* Fill in Call Clear Indicate Control Information Field */
   pIndicate->CtrlInfoLength = sizeof(CI_PCM_CALL_CLEAR_IND);
   pIndicate->PrimitiveCode  = GCCI_CM_PC_PCM_CALL_CLEAR_IND;
   pIndicate->DataInfoLength = 0;
   pIndicate->Selector       = GCCI_COM_SEL_PCM;
   pIndicate->P_CI_ID        = (ULONG)pCall;
   pIndicate->CommonStatus.ReportLocation   = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pIndicate->CommonStatus.ReportCompCode   = CompCode;
   pIndicate->CommonStatus.Reserved         = 0;
   pIndicate->CommonStatus.OrigLocation     = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pIndicate->CommonStatus.OrigCompCode     = CompCode;
   pIndicate->CommonStatus.SysLogID         = 0L;

   if ( pCall->U_CI_ID == 0L )    /* if U_CI_ID is not yet known */
      {
         pIndicate->IDType     = GCCI_COM_ID_TYPE_U_SAP_ID;
         pIndicate->Identifier = pPCMCB->CMSAP.U_SAP_ID;
      }
   else   /* U_CI_ID is known */
      {
         pIndicate->IDType     = GCCI_COM_ID_TYPE_U_CI_ID;
         pIndicate->Identifier = pCall->U_CI_ID;
      }

   if ( CompCode == CCI_NO_ERROR )
      pIndicate->CommonStatus.CompCodeCategory = GCCI_COM_CMP_CAT_REQ_SUCCESSFUL;
   else  /* CompCode is an error */
      pIndicate->CommonStatus.CompCodeCategory = GCCI_COM_CMP_CAT_REQ_REJECTED;

   /* Call Connection Manager with Call Clear Indicate primitive */
   RetCode = pPCMCB->CMSAP.Indicate( pIndicate );

   if ( RetCode == CCI_NO_ERROR )
      {
         free( pIndicate );
         return SUCCESS;
      }
   else  /* Error returned on Call Clear Indicate */
      {
         free( pIndicate );
         return FAILURE;
      }
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : CallClearResponse                                         */
/*                                                                           */
/*   Description : This routine processes a Call Clear response primitive.   */
/*                                                                           */
/*****************************************************************************/

VOID near CallClearResponse( CI_PCM_CALL_CLEAR_RSP *pResponse )
{
   PCALL pCall;

   pCall = (PCALL)pResponse->P_CI_ID;

   TerminateCall( pCall );
   DeleteCall( pCall );
      return;
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : CallConnectedIndicate                                     */
/*                                                                           */
/*   Description : This routine sends a Call Connected indicate primitive    */
/*                 to the Connection Manager.                                */
/*                                                                           */
/*****************************************************************************/

BOOL near CallConnectedIndicate( PCALL pCall, USHORT CompCode )
{
   CI_PCM_CALL_CNTD_IND *pIndicate;
   SF_CC_BEAR_CAP       *pBearer;
   PPCMCB pPCMCB;
   USHORT RetCode;

   pPCMCB = pCall->pPCMCB;

   /* Get memory for Call Connected Indicate primitive */
   pIndicate = malloc(sizeof(CI_PCM_CALL_CNTD_IND)+sizeof(SF_CC_BEAR_CAP));
   if ( pIndicate == 0 )
         return FAILURE;

   /* Fill in Call Connected Indicate Control Information Field */
   pIndicate->CtrlInfoLength = sizeof(CI_PCM_CALL_CNTD_IND);
   pIndicate->PrimitiveCode  = GCCI_CM_PC_PCM_CALL_CNTD_IND;
   pIndicate->DataInfoLength = sizeof(SF_CC_BEAR_CAP);
   pIndicate->Selector       = GCCI_COM_SEL_PCM;
   pIndicate->IDType         = GCCI_COM_ID_TYPE_U_CI_ID;
   pIndicate->U_CI_ID        = pCall->U_CI_ID;

   /* Fill in Bearer Capability Subfield */
   SELECTOROF(pBearer)    = SELECTOROF(pIndicate);
   OFFSETOF(pBearer)      = OFFSETOF(pIndicate)+sizeof(CI_PCM_CALL_CNTD_IND);
   pBearer->SFLength      = sizeof(SF_CC_BEAR_CAP);
   pBearer->SFIdentifier  = SF_SFID_CCITT;
   pBearer->SFType        = SF_CC_TYPE_BEAR_CAP;
   pBearer->SFCallType    = SF_CALL_TYPE_DATA_CIRCUIT;
   pBearer->SFUserRate    = SF_USER_RATE_IN_VALUE_FIELD;
   pBearer->SFUserRateVal = 0;


   /* Call Connection Manager with Call Connected Indicate primitive */
   RetCode = pPCMCB->CMSAP.Indicate( pIndicate );

   if ( RetCode == CCI_NO_ERROR )
      {
         free( pIndicate );
         return SUCCESS;
      }
   else  /* Error returned on Call Connected Indicate */
      {
         free( pIndicate );
         return FAILURE;
      }
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : CallStatusIndicate                                        */
/*                                                                           */
/*   Description : This routine sends a Call Status indicate primitive to    */
/*                 the Connection Manager.                                   */
/*                                                                           */
/*****************************************************************************/

BOOL near CallStatusIndicate( PCALL pCall, UCHAR Status, UCHAR DataLength, PUCHAR pData )
{
   CI_PCM_CALL_STAT_IND *pIndicate;
   SF_CS_CALL_STATUS    *pStatus;
   SF_CC_DISPLAY        *pDisplay;
   PPCMCB pPCMCB;
   USHORT DispSFLength, RetCode;

   pPCMCB = pCall->pPCMCB;

   if ( DataLength == 0 )
      DispSFLength = 0;
   else
      DispSFLength = sizeof(SF_CC_DISPLAY) + DataLength - 1;

   /* Get memory for Call Status Indicate primitive */
   pIndicate = malloc( sizeof(CI_PCM_CALL_STAT_IND) +
                           sizeof(SF_CS_CALL_STATUS) + DispSFLength );
   if ( pIndicate == 0 )
         return FAILURE;

   /* Fill in Call Status Indicate Control Information Field */
   pIndicate->CtrlInfoLength = sizeof(CI_PCM_CALL_STAT_IND);
   pIndicate->PrimitiveCode  = GCCI_CM_PC_PCM_CALL_STAT_IND;
   pIndicate->DataInfoLength = sizeof(SF_CS_CALL_STATUS)+DispSFLength;
   pIndicate->Selector       = GCCI_COM_SEL_PCM;
   pIndicate->IDType         = GCCI_COM_ID_TYPE_U_CI_ID;
   pIndicate->U_CI_ID        = pCall->U_CI_ID;
   pIndicate->CommonStatus.CompCodeCategory = GCCI_COM_CMP_CAT_REQ_SUCCESSFUL;
   pIndicate->CommonStatus.ReportLocation   = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pIndicate->CommonStatus.ReportCompCode   = CCI_NO_ERROR;
   pIndicate->CommonStatus.Reserved         = 0;
   pIndicate->CommonStatus.OrigLocation     = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pIndicate->CommonStatus.OrigCompCode     = CCI_NO_ERROR;
   pIndicate->CommonStatus.SysLogID         = 0L;

   /* Fill in Call Status subfield */
   SELECTOROF(pStatus)       = SELECTOROF(pIndicate);
   OFFSETOF(pStatus)         = OFFSETOF(pIndicate)+sizeof(CI_PCM_CALL_STAT_IND);
   pStatus->SFLength         = sizeof(SF_CS_CALL_STATUS);
   pStatus->SFIdentifier     = SF_SFID_CALL_CONTROL;
   pStatus->SFType           = SF_CS_TYPE_CALL_STATUS;
   pStatus->SFCallStatus     = Status;
   pStatus->SFReserved1      = 0;

   if ( DataLength != 0 )
      {
         /* Fill in Display subfield */
         SELECTOROF(pDisplay)      = SELECTOROF(pIndicate);
         OFFSETOF(pDisplay)        = OFFSETOF(pStatus)+sizeof(SF_CS_CALL_STATUS);
         pDisplay->SFLength        = DispSFLength;
         pDisplay->SFIdentifier    = SF_SFID_CCITT;
         pDisplay->SFType          = SF_CC_TYPE_DISPLAY;
         pDisplay->SFDisplayLength = DataLength;
         strncpy( pDisplay->SFDisplay, pData, DataLength );
      }

   /* Call Connection Manager with Call Status Indicate primitive */
   RetCode = pPCMCB->CMSAP.Indicate( pIndicate );

   if ( RetCode == CCI_NO_ERROR )
      {
         free( pIndicate );
         return SUCCESS;
      }
   else  /* Error returned on Call Status Indicate */
      {
         free( pIndicate );
         return FAILURE;
      }
}


