/*****************************************************************************/
/*                                                                           */
/*   Component   : Sample Port Connection Manager (PCM)                      */
/*                                                                           */
/*   Module      : PCMINIT.C                                                 */
/*                                                                           */
/*   Description : This module contains the routines for starting and        */
/*                 stopping the PCM.  This module also contains the          */
/*                 GCCI routines for activating and deactivating SAPs.       */
/*                                                                           */
/*   Property of IBM.                                                        */
/*   This sample code is available without warranty upon request.            */
/*                                                                           */
/*****************************************************************************/

#define INCL_DOS
#include <os2.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "gcci_com.h"
#include "gcci_cm.h"
#include "sf.h"
#include "cci_rc.h"
#include "pcmdef.h"
#include "pcmproto.h"

BOOL      DLLStarted = FALSE;
PPCMCB    pPCMCBList = NULL;  /* Anchor of linked list of PCM control blocks */

/*****************************************************************************/
/*                                                                           */
/*   Routine     : StartPCM                                                  */
/*                                                                           */
/*   Description : This routine is called by the Module Enabler to start     */
/*                 one or more PCMs.                                         */
/*                                                                           */
/*                 If no PCMs are currently started, then the DLL is         */
/*                 started first.  The requested PCMs are then started.      */
/*                 If no PCM was successfully started, then the DLL is       */
/*                 shutdown.                                                 */
/*                                                                           */
/*****************************************************************************/

USHORT _export _loadds pascal StartPCM( GCCI_COM_MER_INITPCM *pInitCB )
{
   USHORT RetCode, x;

   if ( !DLLStarted )
      {
         RetCode = StartupDLL();
         if ( RetCode == CCI_NO_ERROR )
            {
               DLLStarted = TRUE;
            }
         else  /* Error trying to start the DLL */
            {
               return RetCode;
            }
      }

   for ( x=0; x<pInitCB->num_PCMNames; x++ )
      {
         StartupPCM( pInitCB->PCMNames[x], pInitCB );
      }

   if ( pPCMCBList == NULL )
      return CCI_LACK_OF_RESOURCES;
   else
      return CCI_NO_ERROR;
   return CCI_NO_ERROR;
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : StartupDLL                                                */
/*                                                                           */
/*   Description : This routine performs the global tasks that are           */
/*                 necessary to bring the DLL up to an operational state.    */
/*                 These tasks are needed by all PCMs, but are independent   */
/*                 of any particular PCM.                                    */
/*                                                                           */
/*****************************************************************************/

USHORT near StartupDLL()
{
   USHORT RetCode, DosRC;

   RetCode = CCI_NO_ERROR;

   /* Add ExitDLL routine to exit list */
   if ( DosRC = DosExitList( EXLST_ADD | EXLST_ORDER, ExitDLL ) )
         RetCode = CCI_LACK_OF_RESOURCES;

   return RetCode;
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : ExitDLL                                                   */
/*                                                                           */
/*   Description : This is the exit list routine that is executed if the     */
/*                 Moduler Enabler process is abnormally stopped.  This      */
/*                 routine shuts down all active PCMs and releases all       */
/*                 global resources.                                         */
/*                                                                           */
/*****************************************************************************/

VOID _loadds pascal far ExitDLL( USHORT ExitType )
{
   /* DLL is no longer started */
   DLLStarted = FALSE;

   while ( pPCMCBList != NULL )  /* For each PCMCB */
      {
         /* Shutdown the PCM if it is active */
         if ( pPCMCBList->Status != STATUS_DEACTIVATED )
            ShutdownPCM( pPCMCBList->PCMName );

         /* Free the PCMCB */
         DeletePCMCB( pPCMCBList );
      }

   /* Exit this routine */
   DosExitList( EXLST_EXIT, ExitDLL );
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : StartupPCM                                                */
/*                                                                           */
/*   Description : This routine starts a new PCM.  If the requested PCM      */
/*                 is already started, then the PCM is not started again.    */
/*                                                                           */
/*                 The tasks performed by this routine are:                  */
/*                   - Get configuration data from Module Enabler            */
/*                   - Create a new PCM control block                        */
/*                   - Process the PCM's configuration data                  */
/*                   - Register the PCM to the Module Enabler                */
/*                                                                           */
/*                 This routine does not pass back a return code.            */
/*                 If the PCM is successfully started, then the PCM is       */
/*                 registered with the Moduler Enabler and a PCMCB is        */
/*                 added to the list of PCMCBs.  If a PCM can not be         */
/*                 started, the PCM is not registered with the Module        */
/*                 Enabler and will not be usable for remote LAN access.     */
/*                                                                           */
/*****************************************************************************/

VOID near StartupPCM( PUCHAR pPCMName, GCCI_COM_MER_INITPCM *pInitCB )
{
   GCCI_COM_CI_GET_CONFIG GetConfig;
   GCCI_COM_CI_REG_MODULE RegModule;
   PPCMCB pPCMCB;
   USHORT RetCode;

   pPCMCB = CreatePCMCB( pPCMName );

   if ( pPCMCB != NULL )
   {
      /* Put PCM Name in GetConfig control information */
      if ( !strncpy( GetConfig.PCMName, pPCMName, 16 ) )
         {
            /* Setup GetConfig control information */
            GetConfig.CtrlInfoLength = sizeof(GCCI_COM_CI_GET_CONFIG);
            GetConfig.PrimitiveCode  = GCCI_COM_PC_GET_CONFIG;
            GetConfig.DataInfoLength = 0;
            GetConfig.Selector       = 0;
            GetConfig.IDType         = GCCI_COM_ID_TYPE_NONE;
            GetConfig.Identifier     = 0L;
   
            /* Call Module Enabler to get config data */
            RetCode = pInitCB->MERRegEP(&GetConfig);
            if ( RetCode == CCI_NO_ERROR )
               {
                  /* Process configuration data */
                   
                     /* Put PCM Name in RegModule control information */
                     if ( !strncpy( RegModule.PCMName, pPCMName, 16 ) )
                        {
                           /* Setup RegModule control information */
                           RegModule.PrimitiveCode  = GCCI_COM_PC_REG_MODULE;
                           RegModule.DataInfoLength = 0;
                           RegModule.Selector       = 0;
                           RegModule.IDType         = GCCI_COM_ID_TYPE_NONE;
                           RegModule.Identifier     = 0L;
                           RegModule.PCMActSapEP    = ProcessGCCI;
                           RegModule.PCM_Version    = 0;
   
                           /* Call Module Enabler to register PCM */
                           pInitCB->MERRegEP( &RegModule );
                        }
               }
          }
       }

    return;
}
/*****************************************************************************/
/*                                                                           */
/*   Routine     : ShutdownPCM                                               */
/*                                                                           */
/*   Description : This routine stops a PCM.  The PCM's control block,       */
/*                 SAPs and all call instances are deleted.                  */
/*                                                                           */
/*****************************************************************************/

VOID near ShutdownPCM( PUCHAR pPCMName )
{
   PPCMCB pPCMCB;
   PCALL pCall;

   /* Find the correct PCM control block */
   for ( pPCMCB = pPCMCBList; !strcmp( pPCMName, pPCMCB->PCMName );
         pPCMCB = pPCMCB->pNextPCMCB )
       if ( pPCMCB->pNextPCMCB == NULL ) return;

   /* Delete the Call list */
   if ( pPCMCB->pCallList != NULL )
      {
         for ( pCall=pPCMCB->pCallList;
               pCall->pNextCall!=NULL;
               pCall=pCall->pNextCall );
              {
                 TerminateCall( pCall );
                 DeleteCall( pCall );
              }
      }

   /* CM SAP is now deactivated */
   pPCMCB->CMSAP.Status = STATUS_DEACTIVATED;
   DeletePCMCB( pPCMCB );

   return;
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : CreatePCMCB                                               */
/*                                                                           */
/*   Description : This routine creates a new PCM control block,             */
/*                 initializes it, and adds it to the end of the list        */
/*                 of PCM control blocks.                                    */
/*                                                                           */
/*****************************************************************************/

PPCMCB near CreatePCMCB( PUCHAR pPCMName )
{
   PPCMCB pPCMCB, pLastPCMCB;

   /* Search for a PCMCB with this PCMName */
   for ( pPCMCB=pPCMCBList; pPCMCB!=NULL; pPCMCB=pPCMCB->pNextPCMCB )
      if ( strcmp( pPCMName, pPCMCB->PCMName ) ) break;

   if ( pPCMCB == NULL )  /* PCMCB does not exist for this PCMName */
      {
         /* Create a new PCM control block */
         pPCMCB = malloc( sizeof(PCMCB) );
         if ( pPCMCB == NULL )
               return NULL;

         /* Put PCM Name in PCM control block */
         if ( strncpy( pPCMCB->PCMName, pPCMName, 16 ) )
            {
               free( pPCMCB );
               return NULL;
            }

         /* Initialize the new PCMCB */
         pPCMCB->pNextPCMCB   = NULL;
         pPCMCB->Status       = STATUS_DEACTIVATED;

         /* Add new PCM control block to PCM control block list */
         if ( pPCMCBList == NULL )
            pPCMCBList = pPCMCB;
         else
            {
               for ( pLastPCMCB=pPCMCBList; pLastPCMCB->pNextPCMCB!=NULL;
                     pLastPCMCB=pLastPCMCB->pNextPCMCB );
               pLastPCMCB->pNextPCMCB = pPCMCB;
            }
      }

   if ( pPCMCB->Status == STATUS_DEACTIVATED )
      {
         /* Initialize the PCM control block */
         pPCMCB->Status           = STATUS_ACTIVATING;

         /* Initialize CM SAP */
         pPCMCB->CMSAP.Status   = STATUS_DEACTIVATED;
         pPCMCB->CMSAP.U_SAP_ID = 0L;
         pPCMCB->CMSAP.Indicate = NULL;
         pPCMCB->CMSAP.Confirm  = NULL;
         pPCMCB->CMSAP.pPCMCB   = pPCMCB;

      }
   else  /* PCM is not deactivated */
         return NULL;

   return pPCMCB;
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : DeletePCMCB                                               */
/*                                                                           */
/*   Description : This routine removes a PCM control block from the PCM     */
/*                 control block list and releases its memory.               */
/*                                                                           */
/*****************************************************************************/

PPCMCB near DeletePCMCB( PPCMCB pPCMCB )
{
   PPCMCB  pPrevPCMCB, pNextPCMCB;

   /* Remove the PCM control block the PCMCB list */
   if ( pPCMCBList == pPCMCB )
      {
         pNextPCMCB = pPCMCBList = pPCMCB->pNextPCMCB;
      }
   else
      {
         for ( pPrevPCMCB=pPCMCBList; pPrevPCMCB->pNextPCMCB!=pPCMCB;
               pPrevPCMCB=pPrevPCMCB->pNextPCMCB );
         pNextPCMCB = pPrevPCMCB->pNextPCMCB = pPCMCB->pNextPCMCB;
      }

   free( pPCMCB );      /* Release memory for the PCMCB structure */

   return pNextPCMCB;
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : ProcessGCCI                                               */
/*                                                                           */
/*   Description : This routine processes all GCCI primitives from the       */
/*                 Connection Manager.  The address of this routine is       */
/*                 returned to the Module Enabler during PCM registration.   */
/*                 Each primitive control block is routed ti\o the           */
/*                 appropriate subroutine for processing.                    */
/*                                                                           */
/*****************************************************************************/

USHORT _loadds pascal ProcessGCCI(PGCCI pGCCI)
{
   switch ( pGCCI->PrimitiveCode )
      {
         case GCCI_COM_PC_ACT_SAP_REQ:
            ActivateSAP( (GCCI_COM_CI_ACT_SAP_REQ *)pGCCI);
            break;
         case GCCI_COM_PC_DEACT_SAP_REQ:
            DeactivateSAP( (GCCI_COM_CI_DEACT_SAP_REQ *)pGCCI);
            break;
         case GCCI_CM_PC_PCM_CALL_SETUP_REQ:
            CallSetupRequest( (CI_PCM_CALL_SETUP_REQ *)pGCCI);                         
            break;
         case GCCI_CM_PC_PCM_CALL_SETUP_RSP:
            CallSetupResponse( (CI_PCM_CALL_SETUP_RSP *)pGCCI);                         
            break;
         case GCCI_CM_PC_PCM_CALL_CLEAR_REQ:
            CallClearRequest( (CI_PCM_CALL_CLEAR_REQ *)pGCCI);                         
            break;
         case GCCI_CM_PC_PCM_CALL_CLEAR_RSP:
            CallClearResponse( (CI_PCM_CALL_CLEAR_RSP *)pGCCI);                         
            break;
         case GCCI_CM_PC_PCM_CALL_CTRL_REQ:
            CallControlRequest( (CI_PCM_CALL_CTRL_REQ *)pGCCI);                         
            break;
         default: 
            return CCI_PRIMITIVE;
            break;
      }

   return CCI_NO_ERROR;
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : ActivateSAP                                               */
/*                                                                           */
/*   Description : This routine processes an Activate SAP request primitive. */
/*                                                                           */
/*                 An Activate SAP request primitive will be received        */
/*                 from the Connection Manager for each PCM that was         */
/*                 registered to the Module Enabler.                         */
/*                                                                           */
/*                 This routine creates a SAP and attaches it to the         */
/*                 correct PCM control block.  The entry points to the       */
/*                 Connection Manager are stored in the SAP.                 */
/*                                                                           */
/*****************************************************************************/

VOID near ActivateSAP( GCCI_COM_CI_ACT_SAP_REQ *pRequest )
{
   GCCI_COM_SF01_ACT_SAP_REQ *pLowerDsp;
   PPCMCB  pPCMCB;

   /* Find the PCM control block for this SAP */
   for ( pPCMCB = pPCMCBList;
         !strcmp( pRequest->PCMName, pPCMCB->PCMName );
         pPCMCB = pPCMCB->pNextPCMCB )
      if ( pPCMCB->pNextPCMCB == NULL )
            return;

   SELECTOROF(pLowerDsp) = SELECTOROF(pRequest);
   OFFSETOF(pLowerDsp)   = OFFSETOF(pRequest)+sizeof(GCCI_COM_CI_ACT_SAP_REQ);

   pPCMCB->Status = STATUS_ACTIVATING;

   if ( pRequest->ServiceType & GCCI_COM_SRV_TYPE_CONNECT_MGR )  
      {
         pPCMCB->CMSAP.U_SAP_ID = pRequest->U_SAP_ID;
         pPCMCB->CMSAP.Indicate = pLowerDsp->SFCMIndEP;
         pPCMCB->CMSAP.Confirm  = pLowerDsp->SFCMCnfEP;

         pPCMCB->CMSAP.Status = STATUS_ACTIVATING;

         if (0L == 1L)  /* Note: If no ports/line/channels, etc.are available,    */
            {           /*       then this must be indicated to Connection Manger */
               pPCMCB->CMSAP.Status = STATUS_DEACTIVATED;

               /* Send unsuccessful Activate SAP Confirm */
               CMActivateSAPConfirm( pPCMCB, CCI_RESOURCE_NOT_AVAILABLE );
            }
         else
            /* Send successful Activate SAP Confirm */
            CMActivateSAPConfirm( pPCMCB, CCI_NO_ERROR );
      }
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : CMActivateSAPConfirm                                      */
/*                                                                           */
/*   Description : This routine sends an Activate SAP confrim primitive      */
/*                 with the specified completion code to the Connection      */
/*                 Manager.                                                  */
/*                                                                           */
/*                 This primitive includes subfields that informs the        */
/*                 Connection Manager which resources this PCM manages.      */
/*                                                                           */
/*****************************************************************************/

VOID near CMActivateSAPConfirm( PPCMCB pPCMCB, USHORT CompCode )
{
   GCCI_COM_CI_ACT_SAP_CNF    *pConfirm;
   GCCI_COM_SF01V_ACT_SAP_CNF *pPCMInfo;
   GCCI_COM_SF02_ACT_SAP_CNF  *pUpperDsp;
   VCHANNELINFO               *pChannelInfo;
   USHORT ConfirmSize, PCMInfoSize;
   USHORT PortCnt, RetCode, x;


   /* Calculate size of Activate SAP Confirm primitive */

   PCMInfoSize = sizeof(GCCI_COM_SF01V_ACT_SAP_CNF);
                 
   ConfirmSize = sizeof(GCCI_COM_CI_ACT_SAP_CNF)   +
                 sizeof(GCCI_COM_SF02_ACT_SAP_CNF) +
                 PCMInfoSize;

   /* Get memory for Activate SAP Confirm primitive */
   pConfirm = malloc( ConfirmSize );
   if ( pConfirm == 0 )
         return;

   /* Fill in Activate SAP Confirm Control Information Field */
   pConfirm->CtrlInfoLength = sizeof(GCCI_COM_CI_ACT_SAP_CNF);
   pConfirm->PrimitiveCode  = GCCI_COM_PC_ACT_SAP_CNF;
   pConfirm->DataInfoLength = PCMInfoSize + sizeof(GCCI_COM_SF02_ACT_SAP_CNF);
   pConfirm->Selector       = GCCI_COM_SEL_PCM;
   pConfirm->IDType         = GCCI_COM_ID_TYPE_U_SAP_ID;
   pConfirm->U_SAP_ID       = pPCMCB->CMSAP.U_SAP_ID;
   pConfirm->P_SAP_ID       = (DWORD)(&(pPCMCB->CMSAP));
   for ( x=0; x<16; x++ ) pConfirm->Reserved1[x] = 0;
   pConfirm->CommonStatus.ReportLocation   = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pConfirm->CommonStatus.ReportCompCode   = CompCode;
   pConfirm->CommonStatus.Reserved         = 0;
   pConfirm->CommonStatus.OrigLocation     = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pConfirm->CommonStatus.OrigCompCode     = CompCode;
   pConfirm->CommonStatus.SysLogID         = 0L;

   if ( CompCode == CCI_NO_ERROR )
      pConfirm->CommonStatus.CompCodeCategory = GCCI_COM_CMP_CAT_REQ_SUCCESSFUL;
   else  /* CompCode is an error */
      pConfirm->CommonStatus.CompCodeCategory = GCCI_COM_CMP_CAT_REQ_REJECTED;

   /* Fill in PCM Info Subfield */
   SELECTOROF(pPCMInfo)    = SELECTOROF(pConfirm);
   OFFSETOF(pPCMInfo)      = OFFSETOF(pConfirm) + sizeof(GCCI_COM_CI_ACT_SAP_CNF);
   pPCMInfo->SFLength      = PCMInfoSize;
   pPCMInfo->SFIdentifier  = GCCI_COM_SFID_PRM_SPECIFIC;
   pPCMInfo->SFType        = GCCI_COM_SF_PCM_INFORMATION;
   pPCMInfo->NetworkType   = GCCI_COM_NET_TYPE_VSERIES;
   pPCMInfo->LocalNumCount = (BYTE)PortCnt;
   pPCMInfo->BChannelsDef  = PortCnt;

   /* Fill in Channel Information */
   pChannelInfo = pPCMInfo->ChanInfoArr;

   /* Fill in Upper Dispatch Subfield */
   SELECTOROF(pUpperDsp)   = SELECTOROF(pConfirm);
   OFFSETOF(pUpperDsp)     = OFFSETOF(pPCMInfo) + PCMInfoSize;
   pUpperDsp->SFLength     = sizeof(GCCI_COM_SF02_ACT_SAP_CNF);
   pUpperDsp->SFIdentifier = GCCI_COM_SFID_PRM_SPECIFIC;
   pUpperDsp->SFType       = GCCI_COM_SF_UPPER_DISPATCH;
   pUpperDsp->SFCCReqEP    = ProcessGCCI;
   pUpperDsp->SFCCRspEP    = ProcessGCCI;                                      
   pUpperDsp->SFRASReqEP   = ProcessGCCI;                                      
   pUpperDsp->SFRASRspEP   = ProcessGCCI;                                      

   /* Call Connection Manager with Activate SAP Confirm primitive */
   RetCode = pPCMCB->CMSAP.Confirm( pConfirm );

   free( pConfirm );
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : DeactivateSAP                                             */
/*                                                                           */
/*   Description : This routine processes a Deactivate SAP request           */
/*                 primitive.                                                */
/*                                                                           */
/*****************************************************************************/

VOID near DeactivateSAP( GCCI_COM_CI_DEACT_SAP_REQ *pRequest )
{
   PPCMCB pPCMCB;
   PSAP   pSAP;

   pSAP   = (PSAP)pRequest->P_SAP_ID;
   pPCMCB = pSAP->pPCMCB;

   /* This PCM is now deactivating */
   pPCMCB->Status = STATUS_DEACTIVATING;

    if ( pSAP == &(pPCMCB->CMSAP) )  /* Deactivating CM SAP */
      {
         /* This SAP is now deactivating */
         pSAP->Status = STATUS_DEACTIVATING;
      }
}


/*****************************************************************************/
/*                                                                           */
/*   Routine     : DeactivateSAPConfirm                                      */
/*                                                                           */
/*   Description : This routine sends a Deactivate SAP confirm primitive     */
/*                 to the Connection Manager.                                */
/*                                                                           */
/*****************************************************************************/

VOID near DeactivateSAPConfirm( PSAP pSAP )
{
   GCCI_COM_CI_DEACT_SAP_CNF *pConfirm;
   USHORT RetCode;

   /* Get memory for Deactivate SAP Confirm primitive */
   pConfirm = malloc( sizeof(GCCI_COM_CI_DEACT_SAP_CNF) );
   if ( pConfirm == 0 )
      {
         return;
      }

   /* Fill in Deactivate SAP Confirm Control Information Field */
   pConfirm->CtrlInfoLength = sizeof(GCCI_COM_CI_DEACT_SAP_CNF);
   pConfirm->PrimitiveCode  = GCCI_COM_PC_DEACT_SAP_CNF;
   pConfirm->DataInfoLength = 0;
   pConfirm->Selector       = GCCI_COM_SEL_PCM;
   pConfirm->IDType         = GCCI_COM_ID_TYPE_U_SAP_ID;
   pConfirm->U_SAP_ID       = pSAP->U_SAP_ID;
   pConfirm->CommonStatus.CompCodeCategory = GCCI_COM_CMP_CAT_REQ_SUCCESSFUL;
   pConfirm->CommonStatus.ReportLocation   = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pConfirm->CommonStatus.ReportCompCode   = CCI_NO_ERROR;
   pConfirm->CommonStatus.Reserved         = 0;
   pConfirm->CommonStatus.OrigLocation     = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pConfirm->CommonStatus.OrigCompCode     = CCI_NO_ERROR;
   pConfirm->CommonStatus.SysLogID         = 0L;

   /* Call CM with Deactivate SAP Confirm primitive */
   RetCode = pSAP->Confirm( pConfirm );

   free( pConfirm );
}



/*****************************************************************************/
/*                                                                           */
/*   Routine     : DeactivateSAPIndicate                                     */
/*                                                                           */
/*   Description : This routine sends a Deactivate SAP indicate primitive    */
/*                 to the CM.                                                */
/*                                                                           */
/*****************************************************************************/

BOOL near DeactivateSAPIndicate( PSAP pSAP, USHORT CompCode )
{
   GCCI_COM_CI_DEACT_SAP_IND *pIndicate;
   USHORT RetCode;

   /* Get memory for Deactivate SAP Indicate primitive */
   pIndicate = malloc( sizeof(GCCI_COM_CI_DEACT_SAP_IND) );
   if ( pIndicate == 0 )
      {
         return FAILURE;
      }

   /* Fill in Call Clear Indicate Control Information Field */
   pIndicate->CtrlInfoLength = sizeof(GCCI_COM_CI_DEACT_SAP_IND);
   pIndicate->PrimitiveCode  = GCCI_COM_PC_DEACT_SAP_IND;
   pIndicate->DataInfoLength = 0;
   pIndicate->Selector       = GCCI_COM_SEL_PCM;
   pIndicate->IDType         = GCCI_COM_ID_TYPE_U_SAP_ID;
   pIndicate->U_SAP_ID       = pSAP->U_SAP_ID;
   pIndicate->CommonStatus.ReportLocation   = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pIndicate->CommonStatus.ReportCompCode   = CompCode;
   pIndicate->CommonStatus.Reserved         = 0;
   pIndicate->CommonStatus.OrigLocation     = GCCI_COM_LOC_LOCAL | GCCI_COM_LAYER_ID_PCM;
   pIndicate->CommonStatus.OrigCompCode     = CompCode;
   pIndicate->CommonStatus.SysLogID         = 0L;

   if ( CompCode == CCI_NO_ERROR )
      pIndicate->CommonStatus.CompCodeCategory = GCCI_COM_CMP_CAT_REQ_SUCCESSFUL;
   else  /* CompCode is an error */
      pIndicate->CommonStatus.CompCodeCategory = GCCI_COM_CMP_CAT_REQ_REJECTED;

   /* Call CM with Deactivate SAP Indicate primitive */
   RetCode = pSAP->Indicate( pIndicate );

   if ( RetCode == CCI_NO_ERROR )
      {
         free( pIndicate );
         return SUCCESS;
      }
   else  /* Error returned on Deactivate SAP Indicate */
      {
         free( pIndicate );
         return FAILURE;
      }
}


